// Code generated by protoc-gen-go. DO NOT EDIT.
// source: metrics.proto

package soong_metrics_proto

import (
	fmt "fmt"
	proto "github.com/golang/protobuf/proto"
	math "math"
)

// Reference imports to suppress errors if they are not otherwise used.
var _ = proto.Marshal
var _ = fmt.Errorf
var _ = math.Inf

// This is a compile-time assertion to ensure that this generated file
// is compatible with the proto package it is being compiled against.
// A compilation error at this line likely means your copy of the
// proto package needs to be updated.
const _ = proto.ProtoPackageIsVersion3 // please upgrade the proto package

type MetricsBase_BuildVariant int32

const (
	MetricsBase_USER      MetricsBase_BuildVariant = 0
	MetricsBase_USERDEBUG MetricsBase_BuildVariant = 1
	MetricsBase_ENG       MetricsBase_BuildVariant = 2
)

var MetricsBase_BuildVariant_name = map[int32]string{
	0: "USER",
	1: "USERDEBUG",
	2: "ENG",
}

var MetricsBase_BuildVariant_value = map[string]int32{
	"USER":      0,
	"USERDEBUG": 1,
	"ENG":       2,
}

func (x MetricsBase_BuildVariant) Enum() *MetricsBase_BuildVariant {
	p := new(MetricsBase_BuildVariant)
	*p = x
	return p
}

func (x MetricsBase_BuildVariant) String() string {
	return proto.EnumName(MetricsBase_BuildVariant_name, int32(x))
}

func (x *MetricsBase_BuildVariant) UnmarshalJSON(data []byte) error {
	value, err := proto.UnmarshalJSONEnum(MetricsBase_BuildVariant_value, data, "MetricsBase_BuildVariant")
	if err != nil {
		return err
	}
	*x = MetricsBase_BuildVariant(value)
	return nil
}

func (MetricsBase_BuildVariant) EnumDescriptor() ([]byte, []int) {
	return fileDescriptor_6039342a2ba47b72, []int{0, 0}
}

type MetricsBase_Arch int32

const (
	MetricsBase_UNKNOWN MetricsBase_Arch = 0
	MetricsBase_ARM     MetricsBase_Arch = 1
	MetricsBase_ARM64   MetricsBase_Arch = 2
	MetricsBase_X86     MetricsBase_Arch = 3
	MetricsBase_X86_64  MetricsBase_Arch = 4
)

var MetricsBase_Arch_name = map[int32]string{
	0: "UNKNOWN",
	1: "ARM",
	2: "ARM64",
	3: "X86",
	4: "X86_64",
}

var MetricsBase_Arch_value = map[string]int32{
	"UNKNOWN": 0,
	"ARM":     1,
	"ARM64":   2,
	"X86":     3,
	"X86_64":  4,
}

func (x MetricsBase_Arch) Enum() *MetricsBase_Arch {
	p := new(MetricsBase_Arch)
	*p = x
	return p
}

func (x MetricsBase_Arch) String() string {
	return proto.EnumName(MetricsBase_Arch_name, int32(x))
}

func (x *MetricsBase_Arch) UnmarshalJSON(data []byte) error {
	value, err := proto.UnmarshalJSONEnum(MetricsBase_Arch_value, data, "MetricsBase_Arch")
	if err != nil {
		return err
	}
	*x = MetricsBase_Arch(value)
	return nil
}

func (MetricsBase_Arch) EnumDescriptor() ([]byte, []int) {
	return fileDescriptor_6039342a2ba47b72, []int{0, 1}
}

type ModuleTypeInfo_BuildSystem int32

const (
	ModuleTypeInfo_UNKNOWN ModuleTypeInfo_BuildSystem = 0
	ModuleTypeInfo_SOONG   ModuleTypeInfo_BuildSystem = 1
	ModuleTypeInfo_MAKE    ModuleTypeInfo_BuildSystem = 2
)

var ModuleTypeInfo_BuildSystem_name = map[int32]string{
	0: "UNKNOWN",
	1: "SOONG",
	2: "MAKE",
}

var ModuleTypeInfo_BuildSystem_value = map[string]int32{
	"UNKNOWN": 0,
	"SOONG":   1,
	"MAKE":    2,
}

func (x ModuleTypeInfo_BuildSystem) Enum() *ModuleTypeInfo_BuildSystem {
	p := new(ModuleTypeInfo_BuildSystem)
	*p = x
	return p
}

func (x ModuleTypeInfo_BuildSystem) String() string {
	return proto.EnumName(ModuleTypeInfo_BuildSystem_name, int32(x))
}

func (x *ModuleTypeInfo_BuildSystem) UnmarshalJSON(data []byte) error {
	value, err := proto.UnmarshalJSONEnum(ModuleTypeInfo_BuildSystem_value, data, "ModuleTypeInfo_BuildSystem")
	if err != nil {
		return err
	}
	*x = ModuleTypeInfo_BuildSystem(value)
	return nil
}

func (ModuleTypeInfo_BuildSystem) EnumDescriptor() ([]byte, []int) {
	return fileDescriptor_6039342a2ba47b72, []int{4, 0}
}

type MetricsBase struct {
	// Timestamp generated when the build starts.
	BuildDateTimestamp *int64 `protobuf:"varint,1,opt,name=build_date_timestamp,json=buildDateTimestamp" json:"build_date_timestamp,omitempty"`
	// It is usually used to specify the branch name [and release candidate].
	BuildId *string `protobuf:"bytes,2,opt,name=build_id,json=buildId" json:"build_id,omitempty"`
	// The platform version codename, eg. P, Q, REL.
	PlatformVersionCodename *string `protobuf:"bytes,3,opt,name=platform_version_codename,json=platformVersionCodename" json:"platform_version_codename,omitempty"`
	// The target product information, eg. aosp_arm.
	TargetProduct *string `protobuf:"bytes,4,opt,name=target_product,json=targetProduct" json:"target_product,omitempty"`
	// The target build variant information, eg. eng.
	TargetBuildVariant *MetricsBase_BuildVariant `protobuf:"varint,5,opt,name=target_build_variant,json=targetBuildVariant,enum=soong_build_metrics.MetricsBase_BuildVariant,def=2" json:"target_build_variant,omitempty"`
	// The target arch information, eg. arm.
	TargetArch *MetricsBase_Arch `protobuf:"varint,6,opt,name=target_arch,json=targetArch,enum=soong_build_metrics.MetricsBase_Arch,def=0" json:"target_arch,omitempty"`
	// The target arch variant information, eg. armv7-a-neon.
	TargetArchVariant *string `protobuf:"bytes,7,opt,name=target_arch_variant,json=targetArchVariant" json:"target_arch_variant,omitempty"`
	// The target cpu variant information, eg. generic.
	TargetCpuVariant *string `protobuf:"bytes,8,opt,name=target_cpu_variant,json=targetCpuVariant" json:"target_cpu_variant,omitempty"`
	// The host arch information, eg. x86_64.
	HostArch *MetricsBase_Arch `protobuf:"varint,9,opt,name=host_arch,json=hostArch,enum=soong_build_metrics.MetricsBase_Arch,def=0" json:"host_arch,omitempty"`
	// The host 2nd arch information, eg. x86.
	Host_2NdArch *MetricsBase_Arch `protobuf:"varint,10,opt,name=host_2nd_arch,json=host2ndArch,enum=soong_build_metrics.MetricsBase_Arch,def=0" json:"host_2nd_arch,omitempty"`
	// The host os information, eg. linux.
	HostOs *string `protobuf:"bytes,11,opt,name=host_os,json=hostOs" json:"host_os,omitempty"`
	// The host os extra information, eg. Linux-4.17.0-3rodete2-amd64-x86_64-Debian-GNU.
	HostOsExtra *string `protobuf:"bytes,12,opt,name=host_os_extra,json=hostOsExtra" json:"host_os_extra,omitempty"`
	// The host cross os information, eg. windows.
	HostCrossOs *string `protobuf:"bytes,13,opt,name=host_cross_os,json=hostCrossOs" json:"host_cross_os,omitempty"`
	// The host cross arch information, eg. x86.
	HostCrossArch *string `protobuf:"bytes,14,opt,name=host_cross_arch,json=hostCrossArch" json:"host_cross_arch,omitempty"`
	// The host cross 2nd arch information, eg. x86_64.
	HostCross_2NdArch *string `protobuf:"bytes,15,opt,name=host_cross_2nd_arch,json=hostCross2ndArch" json:"host_cross_2nd_arch,omitempty"`
	// The directory for generated built artifacts installation, eg. out.
	OutDir *string `protobuf:"bytes,16,opt,name=out_dir,json=outDir" json:"out_dir,omitempty"`
	// The metrics for calling various tools (microfactory) before Soong_UI starts.
	SetupTools []*PerfInfo `protobuf:"bytes,17,rep,name=setup_tools,json=setupTools" json:"setup_tools,omitempty"`
	// The metrics for calling Kati by multiple times.
	KatiRuns []*PerfInfo `protobuf:"bytes,18,rep,name=kati_runs,json=katiRuns" json:"kati_runs,omitempty"`
	// The metrics for calling Soong.
	SoongRuns []*PerfInfo `protobuf:"bytes,19,rep,name=soong_runs,json=soongRuns" json:"soong_runs,omitempty"`
	// The metrics for calling Ninja.
	NinjaRuns []*PerfInfo `protobuf:"bytes,20,rep,name=ninja_runs,json=ninjaRuns" json:"ninja_runs,omitempty"`
	// The metrics for the whole build
	Total             *PerfInfo          `protobuf:"bytes,21,opt,name=total" json:"total,omitempty"`
	SoongBuildMetrics *SoongBuildMetrics `protobuf:"bytes,22,opt,name=soong_build_metrics,json=soongBuildMetrics" json:"soong_build_metrics,omitempty"`
	BuildConfig       *BuildConfig       `protobuf:"bytes,23,opt,name=build_config,json=buildConfig" json:"build_config,omitempty"`
	// The hostname of the machine.
	Hostname *string `protobuf:"bytes,24,opt,name=hostname" json:"hostname,omitempty"`
	// The system resource information such as total physical memory.
	SystemResourceInfo   *SystemResourceInfo `protobuf:"bytes,25,opt,name=system_resource_info,json=systemResourceInfo" json:"system_resource_info,omitempty"`
	XXX_NoUnkeyedLiteral struct{}            `json:"-"`
	XXX_unrecognized     []byte              `json:"-"`
	XXX_sizecache        int32               `json:"-"`
}

func (m *MetricsBase) Reset()         { *m = MetricsBase{} }
func (m *MetricsBase) String() string { return proto.CompactTextString(m) }
func (*MetricsBase) ProtoMessage()    {}
func (*MetricsBase) Descriptor() ([]byte, []int) {
	return fileDescriptor_6039342a2ba47b72, []int{0}
}

func (m *MetricsBase) XXX_Unmarshal(b []byte) error {
	return xxx_messageInfo_MetricsBase.Unmarshal(m, b)
}
func (m *MetricsBase) XXX_Marshal(b []byte, deterministic bool) ([]byte, error) {
	return xxx_messageInfo_MetricsBase.Marshal(b, m, deterministic)
}
func (m *MetricsBase) XXX_Merge(src proto.Message) {
	xxx_messageInfo_MetricsBase.Merge(m, src)
}
func (m *MetricsBase) XXX_Size() int {
	return xxx_messageInfo_MetricsBase.Size(m)
}
func (m *MetricsBase) XXX_DiscardUnknown() {
	xxx_messageInfo_MetricsBase.DiscardUnknown(m)
}

var xxx_messageInfo_MetricsBase proto.InternalMessageInfo

const Default_MetricsBase_TargetBuildVariant MetricsBase_BuildVariant = MetricsBase_ENG
const Default_MetricsBase_TargetArch MetricsBase_Arch = MetricsBase_UNKNOWN
const Default_MetricsBase_HostArch MetricsBase_Arch = MetricsBase_UNKNOWN
const Default_MetricsBase_Host_2NdArch MetricsBase_Arch = MetricsBase_UNKNOWN

func (m *MetricsBase) GetBuildDateTimestamp() int64 {
	if m != nil && m.BuildDateTimestamp != nil {
		return *m.BuildDateTimestamp
	}
	return 0
}

func (m *MetricsBase) GetBuildId() string {
	if m != nil && m.BuildId != nil {
		return *m.BuildId
	}
	return ""
}

func (m *MetricsBase) GetPlatformVersionCodename() string {
	if m != nil && m.PlatformVersionCodename != nil {
		return *m.PlatformVersionCodename
	}
	return ""
}

func (m *MetricsBase) GetTargetProduct() string {
	if m != nil && m.TargetProduct != nil {
		return *m.TargetProduct
	}
	return ""
}

func (m *MetricsBase) GetTargetBuildVariant() MetricsBase_BuildVariant {
	if m != nil && m.TargetBuildVariant != nil {
		return *m.TargetBuildVariant
	}
	return Default_MetricsBase_TargetBuildVariant
}

func (m *MetricsBase) GetTargetArch() MetricsBase_Arch {
	if m != nil && m.TargetArch != nil {
		return *m.TargetArch
	}
	return Default_MetricsBase_TargetArch
}

func (m *MetricsBase) GetTargetArchVariant() string {
	if m != nil && m.TargetArchVariant != nil {
		return *m.TargetArchVariant
	}
	return ""
}

func (m *MetricsBase) GetTargetCpuVariant() string {
	if m != nil && m.TargetCpuVariant != nil {
		return *m.TargetCpuVariant
	}
	return ""
}

func (m *MetricsBase) GetHostArch() MetricsBase_Arch {
	if m != nil && m.HostArch != nil {
		return *m.HostArch
	}
	return Default_MetricsBase_HostArch
}

func (m *MetricsBase) GetHost_2NdArch() MetricsBase_Arch {
	if m != nil && m.Host_2NdArch != nil {
		return *m.Host_2NdArch
	}
	return Default_MetricsBase_Host_2NdArch
}

func (m *MetricsBase) GetHostOs() string {
	if m != nil && m.HostOs != nil {
		return *m.HostOs
	}
	return ""
}

func (m *MetricsBase) GetHostOsExtra() string {
	if m != nil && m.HostOsExtra != nil {
		return *m.HostOsExtra
	}
	return ""
}

func (m *MetricsBase) GetHostCrossOs() string {
	if m != nil && m.HostCrossOs != nil {
		return *m.HostCrossOs
	}
	return ""
}

func (m *MetricsBase) GetHostCrossArch() string {
	if m != nil && m.HostCrossArch != nil {
		return *m.HostCrossArch
	}
	return ""
}

func (m *MetricsBase) GetHostCross_2NdArch() string {
	if m != nil && m.HostCross_2NdArch != nil {
		return *m.HostCross_2NdArch
	}
	return ""
}

func (m *MetricsBase) GetOutDir() string {
	if m != nil && m.OutDir != nil {
		return *m.OutDir
	}
	return ""
}

func (m *MetricsBase) GetSetupTools() []*PerfInfo {
	if m != nil {
		return m.SetupTools
	}
	return nil
}

func (m *MetricsBase) GetKatiRuns() []*PerfInfo {
	if m != nil {
		return m.KatiRuns
	}
	return nil
}

func (m *MetricsBase) GetSoongRuns() []*PerfInfo {
	if m != nil {
		return m.SoongRuns
	}
	return nil
}

func (m *MetricsBase) GetNinjaRuns() []*PerfInfo {
	if m != nil {
		return m.NinjaRuns
	}
	return nil
}

func (m *MetricsBase) GetTotal() *PerfInfo {
	if m != nil {
		return m.Total
	}
	return nil
}

func (m *MetricsBase) GetSoongBuildMetrics() *SoongBuildMetrics {
	if m != nil {
		return m.SoongBuildMetrics
	}
	return nil
}

func (m *MetricsBase) GetBuildConfig() *BuildConfig {
	if m != nil {
		return m.BuildConfig
	}
	return nil
}

func (m *MetricsBase) GetHostname() string {
	if m != nil && m.Hostname != nil {
		return *m.Hostname
	}
	return ""
}

func (m *MetricsBase) GetSystemResourceInfo() *SystemResourceInfo {
	if m != nil {
		return m.SystemResourceInfo
	}
	return nil
}

type BuildConfig struct {
	UseGoma              *bool    `protobuf:"varint,1,opt,name=use_goma,json=useGoma" json:"use_goma,omitempty"`
	UseRbe               *bool    `protobuf:"varint,2,opt,name=use_rbe,json=useRbe" json:"use_rbe,omitempty"`
	ForceUseGoma         *bool    `protobuf:"varint,3,opt,name=force_use_goma,json=forceUseGoma" json:"force_use_goma,omitempty"`
	XXX_NoUnkeyedLiteral struct{} `json:"-"`
	XXX_unrecognized     []byte   `json:"-"`
	XXX_sizecache        int32    `json:"-"`
}

func (m *BuildConfig) Reset()         { *m = BuildConfig{} }
func (m *BuildConfig) String() string { return proto.CompactTextString(m) }
func (*BuildConfig) ProtoMessage()    {}
func (*BuildConfig) Descriptor() ([]byte, []int) {
	return fileDescriptor_6039342a2ba47b72, []int{1}
}

func (m *BuildConfig) XXX_Unmarshal(b []byte) error {
	return xxx_messageInfo_BuildConfig.Unmarshal(m, b)
}
func (m *BuildConfig) XXX_Marshal(b []byte, deterministic bool) ([]byte, error) {
	return xxx_messageInfo_BuildConfig.Marshal(b, m, deterministic)
}
func (m *BuildConfig) XXX_Merge(src proto.Message) {
	xxx_messageInfo_BuildConfig.Merge(m, src)
}
func (m *BuildConfig) XXX_Size() int {
	return xxx_messageInfo_BuildConfig.Size(m)
}
func (m *BuildConfig) XXX_DiscardUnknown() {
	xxx_messageInfo_BuildConfig.DiscardUnknown(m)
}

var xxx_messageInfo_BuildConfig proto.InternalMessageInfo

func (m *BuildConfig) GetUseGoma() bool {
	if m != nil && m.UseGoma != nil {
		return *m.UseGoma
	}
	return false
}

func (m *BuildConfig) GetUseRbe() bool {
	if m != nil && m.UseRbe != nil {
		return *m.UseRbe
	}
	return false
}

func (m *BuildConfig) GetForceUseGoma() bool {
	if m != nil && m.ForceUseGoma != nil {
		return *m.ForceUseGoma
	}
	return false
}

type SystemResourceInfo struct {
	// The total physical memory in bytes.
	TotalPhysicalMemory *uint64 `protobuf:"varint,1,opt,name=total_physical_memory,json=totalPhysicalMemory" json:"total_physical_memory,omitempty"`
	// The total of available cores for building
	AvailableCpus        *int32   `protobuf:"varint,2,opt,name=available_cpus,json=availableCpus" json:"available_cpus,omitempty"`
	XXX_NoUnkeyedLiteral struct{} `json:"-"`
	XXX_unrecognized     []byte   `json:"-"`
	XXX_sizecache        int32    `json:"-"`
}

func (m *SystemResourceInfo) Reset()         { *m = SystemResourceInfo{} }
func (m *SystemResourceInfo) String() string { return proto.CompactTextString(m) }
func (*SystemResourceInfo) ProtoMessage()    {}
func (*SystemResourceInfo) Descriptor() ([]byte, []int) {
	return fileDescriptor_6039342a2ba47b72, []int{2}
}

func (m *SystemResourceInfo) XXX_Unmarshal(b []byte) error {
	return xxx_messageInfo_SystemResourceInfo.Unmarshal(m, b)
}
func (m *SystemResourceInfo) XXX_Marshal(b []byte, deterministic bool) ([]byte, error) {
	return xxx_messageInfo_SystemResourceInfo.Marshal(b, m, deterministic)
}
func (m *SystemResourceInfo) XXX_Merge(src proto.Message) {
	xxx_messageInfo_SystemResourceInfo.Merge(m, src)
}
func (m *SystemResourceInfo) XXX_Size() int {
	return xxx_messageInfo_SystemResourceInfo.Size(m)
}
func (m *SystemResourceInfo) XXX_DiscardUnknown() {
	xxx_messageInfo_SystemResourceInfo.DiscardUnknown(m)
}

var xxx_messageInfo_SystemResourceInfo proto.InternalMessageInfo

func (m *SystemResourceInfo) GetTotalPhysicalMemory() uint64 {
	if m != nil && m.TotalPhysicalMemory != nil {
		return *m.TotalPhysicalMemory
	}
	return 0
}

func (m *SystemResourceInfo) GetAvailableCpus() int32 {
	if m != nil && m.AvailableCpus != nil {
		return *m.AvailableCpus
	}
	return 0
}

type PerfInfo struct {
	// The description for the phase/action/part while the tool running.
	Desc *string `protobuf:"bytes,1,opt,name=desc" json:"desc,omitempty"`
	// The name for the running phase/action/part.
	Name *string `protobuf:"bytes,2,opt,name=name" json:"name,omitempty"`
	// The absolute start time.
	// The number of nanoseconds elapsed since January 1, 1970 UTC.
	StartTime *uint64 `protobuf:"varint,3,opt,name=start_time,json=startTime" json:"start_time,omitempty"`
	// The real running time.
	// The number of nanoseconds elapsed since start_time.
	RealTime *uint64 `protobuf:"varint,4,opt,name=real_time,json=realTime" json:"real_time,omitempty"`
	// The number of MB for memory use.
	MemoryUse            *uint64  `protobuf:"varint,5,opt,name=memory_use,json=memoryUse" json:"memory_use,omitempty"`
	XXX_NoUnkeyedLiteral struct{} `json:"-"`
	XXX_unrecognized     []byte   `json:"-"`
	XXX_sizecache        int32    `json:"-"`
}

func (m *PerfInfo) Reset()         { *m = PerfInfo{} }
func (m *PerfInfo) String() string { return proto.CompactTextString(m) }
func (*PerfInfo) ProtoMessage()    {}
func (*PerfInfo) Descriptor() ([]byte, []int) {
	return fileDescriptor_6039342a2ba47b72, []int{3}
}

func (m *PerfInfo) XXX_Unmarshal(b []byte) error {
	return xxx_messageInfo_PerfInfo.Unmarshal(m, b)
}
func (m *PerfInfo) XXX_Marshal(b []byte, deterministic bool) ([]byte, error) {
	return xxx_messageInfo_PerfInfo.Marshal(b, m, deterministic)
}
func (m *PerfInfo) XXX_Merge(src proto.Message) {
	xxx_messageInfo_PerfInfo.Merge(m, src)
}
func (m *PerfInfo) XXX_Size() int {
	return xxx_messageInfo_PerfInfo.Size(m)
}
func (m *PerfInfo) XXX_DiscardUnknown() {
	xxx_messageInfo_PerfInfo.DiscardUnknown(m)
}

var xxx_messageInfo_PerfInfo proto.InternalMessageInfo

func (m *PerfInfo) GetDesc() string {
	if m != nil && m.Desc != nil {
		return *m.Desc
	}
	return ""
}

func (m *PerfInfo) GetName() string {
	if m != nil && m.Name != nil {
		return *m.Name
	}
	return ""
}

func (m *PerfInfo) GetStartTime() uint64 {
	if m != nil && m.StartTime != nil {
		return *m.StartTime
	}
	return 0
}

func (m *PerfInfo) GetRealTime() uint64 {
	if m != nil && m.RealTime != nil {
		return *m.RealTime
	}
	return 0
}

func (m *PerfInfo) GetMemoryUse() uint64 {
	if m != nil && m.MemoryUse != nil {
		return *m.MemoryUse
	}
	return 0
}

type ModuleTypeInfo struct {
	// The build system, eg. Soong or Make.
	BuildSystem *ModuleTypeInfo_BuildSystem `protobuf:"varint,1,opt,name=build_system,json=buildSystem,enum=soong_build_metrics.ModuleTypeInfo_BuildSystem,def=0" json:"build_system,omitempty"`
	// The module type, eg. java_library, cc_binary, and etc.
	ModuleType *string `protobuf:"bytes,2,opt,name=module_type,json=moduleType" json:"module_type,omitempty"`
	// The number of logical modules.
	NumOfModules         *uint32  `protobuf:"varint,3,opt,name=num_of_modules,json=numOfModules" json:"num_of_modules,omitempty"`
	XXX_NoUnkeyedLiteral struct{} `json:"-"`
	XXX_unrecognized     []byte   `json:"-"`
	XXX_sizecache        int32    `json:"-"`
}

func (m *ModuleTypeInfo) Reset()         { *m = ModuleTypeInfo{} }
func (m *ModuleTypeInfo) String() string { return proto.CompactTextString(m) }
func (*ModuleTypeInfo) ProtoMessage()    {}
func (*ModuleTypeInfo) Descriptor() ([]byte, []int) {
	return fileDescriptor_6039342a2ba47b72, []int{4}
}

func (m *ModuleTypeInfo) XXX_Unmarshal(b []byte) error {
	return xxx_messageInfo_ModuleTypeInfo.Unmarshal(m, b)
}
func (m *ModuleTypeInfo) XXX_Marshal(b []byte, deterministic bool) ([]byte, error) {
	return xxx_messageInfo_ModuleTypeInfo.Marshal(b, m, deterministic)
}
func (m *ModuleTypeInfo) XXX_Merge(src proto.Message) {
	xxx_messageInfo_ModuleTypeInfo.Merge(m, src)
}
func (m *ModuleTypeInfo) XXX_Size() int {
	return xxx_messageInfo_ModuleTypeInfo.Size(m)
}
func (m *ModuleTypeInfo) XXX_DiscardUnknown() {
	xxx_messageInfo_ModuleTypeInfo.DiscardUnknown(m)
}

var xxx_messageInfo_ModuleTypeInfo proto.InternalMessageInfo

const Default_ModuleTypeInfo_BuildSystem ModuleTypeInfo_BuildSystem = ModuleTypeInfo_UNKNOWN

func (m *ModuleTypeInfo) GetBuildSystem() ModuleTypeInfo_BuildSystem {
	if m != nil && m.BuildSystem != nil {
		return *m.BuildSystem
	}
	return Default_ModuleTypeInfo_BuildSystem
}

func (m *ModuleTypeInfo) GetModuleType() string {
	if m != nil && m.ModuleType != nil {
		return *m.ModuleType
	}
	return ""
}

func (m *ModuleTypeInfo) GetNumOfModules() uint32 {
	if m != nil && m.NumOfModules != nil {
		return *m.NumOfModules
	}
	return 0
}

type CriticalUserJourneyMetrics struct {
	// The name of a critical user journey test.
	Name *string `protobuf:"bytes,1,opt,name=name" json:"name,omitempty"`
	// The metrics produced when running the critical user journey test.
	Metrics              *MetricsBase `protobuf:"bytes,2,opt,name=metrics" json:"metrics,omitempty"`
	XXX_NoUnkeyedLiteral struct{}     `json:"-"`
	XXX_unrecognized     []byte       `json:"-"`
	XXX_sizecache        int32        `json:"-"`
}

func (m *CriticalUserJourneyMetrics) Reset()         { *m = CriticalUserJourneyMetrics{} }
func (m *CriticalUserJourneyMetrics) String() string { return proto.CompactTextString(m) }
func (*CriticalUserJourneyMetrics) ProtoMessage()    {}
func (*CriticalUserJourneyMetrics) Descriptor() ([]byte, []int) {
	return fileDescriptor_6039342a2ba47b72, []int{5}
}

func (m *CriticalUserJourneyMetrics) XXX_Unmarshal(b []byte) error {
	return xxx_messageInfo_CriticalUserJourneyMetrics.Unmarshal(m, b)
}
func (m *CriticalUserJourneyMetrics) XXX_Marshal(b []byte, deterministic bool) ([]byte, error) {
	return xxx_messageInfo_CriticalUserJourneyMetrics.Marshal(b, m, deterministic)
}
func (m *CriticalUserJourneyMetrics) XXX_Merge(src proto.Message) {
	xxx_messageInfo_CriticalUserJourneyMetrics.Merge(m, src)
}
func (m *CriticalUserJourneyMetrics) XXX_Size() int {
	return xxx_messageInfo_CriticalUserJourneyMetrics.Size(m)
}
func (m *CriticalUserJourneyMetrics) XXX_DiscardUnknown() {
	xxx_messageInfo_CriticalUserJourneyMetrics.DiscardUnknown(m)
}

var xxx_messageInfo_CriticalUserJourneyMetrics proto.InternalMessageInfo

func (m *CriticalUserJourneyMetrics) GetName() string {
	if m != nil && m.Name != nil {
		return *m.Name
	}
	return ""
}

func (m *CriticalUserJourneyMetrics) GetMetrics() *MetricsBase {
	if m != nil {
		return m.Metrics
	}
	return nil
}

type CriticalUserJourneysMetrics struct {
	// A set of metrics from a run of the critical user journey tests.
	Cujs                 []*CriticalUserJourneyMetrics `protobuf:"bytes,1,rep,name=cujs" json:"cujs,omitempty"`
	XXX_NoUnkeyedLiteral struct{}                      `json:"-"`
	XXX_unrecognized     []byte                        `json:"-"`
	XXX_sizecache        int32                         `json:"-"`
}

func (m *CriticalUserJourneysMetrics) Reset()         { *m = CriticalUserJourneysMetrics{} }
func (m *CriticalUserJourneysMetrics) String() string { return proto.CompactTextString(m) }
func (*CriticalUserJourneysMetrics) ProtoMessage()    {}
func (*CriticalUserJourneysMetrics) Descriptor() ([]byte, []int) {
	return fileDescriptor_6039342a2ba47b72, []int{6}
}

func (m *CriticalUserJourneysMetrics) XXX_Unmarshal(b []byte) error {
	return xxx_messageInfo_CriticalUserJourneysMetrics.Unmarshal(m, b)
}
func (m *CriticalUserJourneysMetrics) XXX_Marshal(b []byte, deterministic bool) ([]byte, error) {
	return xxx_messageInfo_CriticalUserJourneysMetrics.Marshal(b, m, deterministic)
}
func (m *CriticalUserJourneysMetrics) XXX_Merge(src proto.Message) {
	xxx_messageInfo_CriticalUserJourneysMetrics.Merge(m, src)
}
func (m *CriticalUserJourneysMetrics) XXX_Size() int {
	return xxx_messageInfo_CriticalUserJourneysMetrics.Size(m)
}
func (m *CriticalUserJourneysMetrics) XXX_DiscardUnknown() {
	xxx_messageInfo_CriticalUserJourneysMetrics.DiscardUnknown(m)
}

var xxx_messageInfo_CriticalUserJourneysMetrics proto.InternalMessageInfo

func (m *CriticalUserJourneysMetrics) GetCujs() []*CriticalUserJourneyMetrics {
	if m != nil {
		return m.Cujs
	}
	return nil
}

type SoongBuildMetrics struct {
	// The number of modules handled by soong_build.
	Modules *uint32 `protobuf:"varint,1,opt,name=modules" json:"modules,omitempty"`
	// The total number of variants handled by soong_build.
	Variants *uint32 `protobuf:"varint,2,opt,name=variants" json:"variants,omitempty"`
	// The total number of allocations in soong_build.
	TotalAllocCount *uint64 `protobuf:"varint,3,opt,name=total_alloc_count,json=totalAllocCount" json:"total_alloc_count,omitempty"`
	// The total size of allocations in soong_build in bytes.
	TotalAllocSize *uint64 `protobuf:"varint,4,opt,name=total_alloc_size,json=totalAllocSize" json:"total_alloc_size,omitempty"`
	// The approximate maximum size of the heap in soong_build in bytes.
	MaxHeapSize          *uint64  `protobuf:"varint,5,opt,name=max_heap_size,json=maxHeapSize" json:"max_heap_size,omitempty"`
	XXX_NoUnkeyedLiteral struct{} `json:"-"`
	XXX_unrecognized     []byte   `json:"-"`
	XXX_sizecache        int32    `json:"-"`
}

func (m *SoongBuildMetrics) Reset()         { *m = SoongBuildMetrics{} }
func (m *SoongBuildMetrics) String() string { return proto.CompactTextString(m) }
func (*SoongBuildMetrics) ProtoMessage()    {}
func (*SoongBuildMetrics) Descriptor() ([]byte, []int) {
	return fileDescriptor_6039342a2ba47b72, []int{7}
}

func (m *SoongBuildMetrics) XXX_Unmarshal(b []byte) error {
	return xxx_messageInfo_SoongBuildMetrics.Unmarshal(m, b)
}
func (m *SoongBuildMetrics) XXX_Marshal(b []byte, deterministic bool) ([]byte, error) {
	return xxx_messageInfo_SoongBuildMetrics.Marshal(b, m, deterministic)
}
func (m *SoongBuildMetrics) XXX_Merge(src proto.Message) {
	xxx_messageInfo_SoongBuildMetrics.Merge(m, src)
}
func (m *SoongBuildMetrics) XXX_Size() int {
	return xxx_messageInfo_SoongBuildMetrics.Size(m)
}
func (m *SoongBuildMetrics) XXX_DiscardUnknown() {
	xxx_messageInfo_SoongBuildMetrics.DiscardUnknown(m)
}

var xxx_messageInfo_SoongBuildMetrics proto.InternalMessageInfo

func (m *SoongBuildMetrics) GetModules() uint32 {
	if m != nil && m.Modules != nil {
		return *m.Modules
	}
	return 0
}

func (m *SoongBuildMetrics) GetVariants() uint32 {
	if m != nil && m.Variants != nil {
		return *m.Variants
	}
	return 0
}

func (m *SoongBuildMetrics) GetTotalAllocCount() uint64 {
	if m != nil && m.TotalAllocCount != nil {
		return *m.TotalAllocCount
	}
	return 0
}

func (m *SoongBuildMetrics) GetTotalAllocSize() uint64 {
	if m != nil && m.TotalAllocSize != nil {
		return *m.TotalAllocSize
	}
	return 0
}

func (m *SoongBuildMetrics) GetMaxHeapSize() uint64 {
	if m != nil && m.MaxHeapSize != nil {
		return *m.MaxHeapSize
	}
	return 0
}

func init() {
	proto.RegisterEnum("soong_build_metrics.MetricsBase_BuildVariant", MetricsBase_BuildVariant_name, MetricsBase_BuildVariant_value)
	proto.RegisterEnum("soong_build_metrics.MetricsBase_Arch", MetricsBase_Arch_name, MetricsBase_Arch_value)
	proto.RegisterEnum("soong_build_metrics.ModuleTypeInfo_BuildSystem", ModuleTypeInfo_BuildSystem_name, ModuleTypeInfo_BuildSystem_value)
	proto.RegisterType((*MetricsBase)(nil), "soong_build_metrics.MetricsBase")
	proto.RegisterType((*BuildConfig)(nil), "soong_build_metrics.BuildConfig")
	proto.RegisterType((*SystemResourceInfo)(nil), "soong_build_metrics.SystemResourceInfo")
	proto.RegisterType((*PerfInfo)(nil), "soong_build_metrics.PerfInfo")
	proto.RegisterType((*ModuleTypeInfo)(nil), "soong_build_metrics.ModuleTypeInfo")
	proto.RegisterType((*CriticalUserJourneyMetrics)(nil), "soong_build_metrics.CriticalUserJourneyMetrics")
	proto.RegisterType((*CriticalUserJourneysMetrics)(nil), "soong_build_metrics.CriticalUserJourneysMetrics")
	proto.RegisterType((*SoongBuildMetrics)(nil), "soong_build_metrics.SoongBuildMetrics")
}

func init() {
	proto.RegisterFile("metrics.proto", fileDescriptor_6039342a2ba47b72)
}

var fileDescriptor_6039342a2ba47b72 = []byte{
	// 1130 bytes of a gzipped FileDescriptorProto
	0x1f, 0x8b, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0xff, 0x9c, 0x56, 0x6f, 0x6f, 0xd4, 0x46,
	0x13, 0xc7, 0xc9, 0x25, 0x77, 0x1e, 0xe7, 0x0e, 0x67, 0x13, 0x1e, 0x0c, 0x08, 0x3d, 0x91, 0x55,
	0x68, 0x54, 0x95, 0x80, 0xae, 0x28, 0x42, 0x11, 0xaa, 0x94, 0x1c, 0x11, 0xa5, 0xe8, 0x72, 0x68,
	0xc3, 0x51, 0xda, 0xbe, 0xb0, 0xf6, 0xec, 0xbd, 0xc4, 0xd4, 0xf6, 0x5a, 0xbb, 0x6b, 0xc4, 0xf1,
	0x1d, 0xfa, 0xa9, 0xfa, 0x59, 0xfa, 0x11, 0xfa, 0xbe, 0xda, 0x59, 0xfb, 0xee, 0x28, 0xd7, 0x82,
	0x78, 0x77, 0x9e, 0xdf, 0x9f, 0x9d, 0x19, 0x8f, 0x67, 0x0f, 0xba, 0x39, 0xd7, 0x32, 0x8d, 0xd5,
	0x41, 0x29, 0x85, 0x16, 0x64, 0x47, 0x09, 0x51, 0x5c, 0x44, 0x93, 0x2a, 0xcd, 0x92, 0xa8, 0x86,
	0xc2, 0xbf, 0x3c, 0xf0, 0x86, 0xf6, 0xf7, 0x09, 0x53, 0x9c, 0x3c, 0x80, 0x5d, 0x4b, 0x48, 0x98,
	0xe6, 0x91, 0x4e, 0x73, 0xae, 0x34, 0xcb, 0xcb, 0xc0, 0xd9, 0x73, 0xf6, 0xd7, 0x29, 0x41, 0xec,
	0x09, 0xd3, 0xfc, 0x65, 0x83, 0x90, 0x1b, 0xd0, 0xb1, 0x8a, 0x34, 0x09, 0xd6, 0xf6, 0x9c, 0x7d,
	0x97, 0xb6, 0xf1, 0xf9, 0x59, 0x42, 0x8e, 0xe0, 0x46, 0x99, 0x31, 0x3d, 0x15, 0x32, 0x8f, 0xde,
	0x72, 0xa9, 0x52, 0x51, 0x44, 0xb1, 0x48, 0x78, 0xc1, 0x72, 0x1e, 0xac, 0x23, 0xf7, 0x7a, 0x43,
	0x78, 0x65, 0xf1, 0x41, 0x0d, 0x93, 0x3b, 0xd0, 0xd3, 0x4c, 0x5e, 0x70, 0x1d, 0x95, 0x52, 0x24,
	0x55, 0xac, 0x83, 0x16, 0x0a, 0xba, 0x36, 0xfa, 0xc2, 0x06, 0x49, 0x02, 0xbb, 0x35, 0xcd, 0x26,
	0xf1, 0x96, 0xc9, 0x94, 0x15, 0x3a, 0xd8, 0xd8, 0x73, 0xf6, 0x7b, 0xfd, 0x7b, 0x07, 0x2b, 0x6a,
	0x3e, 0x58, 0xaa, 0xf7, 0xe0, 0xc4, 0x20, 0xaf, 0xac, 0xe8, 0x68, 0xfd, 0xf4, 0xec, 0x29, 0x25,
	0xd6, 0x6f, 0x19, 0x20, 0x23, 0xf0, 0xea, 0x53, 0x98, 0x8c, 0x2f, 0x83, 0x4d, 0x34, 0xbf, 0xf3,
	0x49, 0xf3, 0x63, 0x19, 0x5f, 0x1e, 0xb5, 0xc7, 0x67, 0xcf, 0xcf, 0x46, 0x3f, 0x9d, 0x51, 0xb0,
	0x16, 0x26, 0x48, 0x0e, 0x60, 0x67, 0xc9, 0x70, 0x9e, 0x75, 0x1b, 0x4b, 0xdc, 0x5e, 0x10, 0x9b,
	0x04, 0xbe, 0x85, 0x3a, 0xad, 0x28, 0x2e, 0xab, 0x39, 0xbd, 0x83, 0x74, 0xdf, 0x22, 0x83, 0xb2,
	0x6a, 0xd8, 0xcf, 0xc1, 0xbd, 0x14, 0xaa, 0x4e, 0xd6, 0xfd, 0xa2, 0x64, 0x3b, 0xc6, 0x00, 0x53,
	0xa5, 0xd0, 0x45, 0xb3, 0x7e, 0x91, 0x58, 0x43, 0xf8, 0x22, 0x43, 0xcf, 0x98, 0xf4, 0x8b, 0x04,
	0x3d, 0xaf, 0x43, 0x1b, 0x3d, 0x85, 0x0a, 0x3c, 0xac, 0x61, 0xd3, 0x3c, 0x8e, 0x14, 0x09, 0xeb,
	0xc3, 0x84, 0x8a, 0xf8, 0x3b, 0x2d, 0x59, 0xb0, 0x85, 0xb0, 0x67, 0xe1, 0x53, 0x13, 0x9a, 0x73,
	0x62, 0x29, 0x94, 0x32, 0x16, 0xdd, 0x05, 0x67, 0x60, 0x62, 0x23, 0x45, 0xee, 0xc2, 0xd5, 0x25,
	0x0e, 0xa6, 0xdd, 0xb3, 0xe3, 0x33, 0x67, 0x61, 0x22, 0xf7, 0x60, 0x67, 0x89, 0x37, 0x2f, 0xf1,
	0xaa, 0x6d, 0xec, 0x9c, 0xbb, 0x94, 0xb7, 0xa8, 0x74, 0x94, 0xa4, 0x32, 0xf0, 0x6d, 0xde, 0xa2,
	0xd2, 0x4f, 0x52, 0x49, 0xbe, 0x07, 0x4f, 0x71, 0x5d, 0x95, 0x91, 0x16, 0x22, 0x53, 0xc1, 0xf6,
	0xde, 0xfa, 0xbe, 0xd7, 0xbf, 0xbd, 0xb2, 0x45, 0x2f, 0xb8, 0x9c, 0x3e, 0x2b, 0xa6, 0x82, 0x02,
	0x2a, 0x5e, 0x1a, 0x01, 0x39, 0x02, 0xf7, 0x37, 0xa6, 0xd3, 0x48, 0x56, 0x85, 0x0a, 0xc8, 0xe7,
	0xa8, 0x3b, 0x86, 0x4f, 0xab, 0x42, 0x91, 0xc7, 0x00, 0x96, 0x89, 0xe2, 0x9d, 0xcf, 0x11, 0xbb,
	0x88, 0x36, 0xea, 0x22, 0x2d, 0xde, 0x30, 0xab, 0xde, 0xfd, 0x2c, 0x35, 0x0a, 0x50, 0xfd, 0x1d,
	0x6c, 0x68, 0xa1, 0x59, 0x16, 0x5c, 0xdb, 0x73, 0x3e, 0x2d, 0xb4, 0x5c, 0xf2, 0x0a, 0x56, 0xad,
	0xa2, 0xe0, 0x7f, 0x68, 0x71, 0x77, 0xa5, 0xc5, 0xb9, 0x89, 0xe1, 0x27, 0x59, 0x4f, 0x18, 0xdd,
	0x56, 0xff, 0x0c, 0x91, 0x01, 0x6c, 0x59, 0x55, 0x2c, 0x8a, 0x69, 0x7a, 0x11, 0x5c, 0x47, 0xc3,
	0xbd, 0x95, 0x86, 0x28, 0x1c, 0x20, 0x8f, 0x7a, 0x93, 0xc5, 0x03, 0xb9, 0x09, 0x38, 0xfa, 0xb8,
	0xa2, 0x02, 0x7c, 0xc7, 0xf3, 0x67, 0xf2, 0x33, 0xec, 0xaa, 0x99, 0xd2, 0x3c, 0x8f, 0x24, 0x57,
	0xa2, 0x92, 0x31, 0x8f, 0xd2, 0x62, 0x2a, 0x82, 0x1b, 0x78, 0xd0, 0xd7, 0xab, 0x33, 0x47, 0x01,
	0xad, 0xf9, 0xd8, 0x06, 0xa2, 0x3e, 0x8a, 0x85, 0x0f, 0x60, 0xeb, 0x83, 0x8d, 0xd3, 0x81, 0xd6,
	0xf8, 0xfc, 0x94, 0xfa, 0x57, 0x48, 0x17, 0x5c, 0xf3, 0xeb, 0xc9, 0xe9, 0xc9, 0xf8, 0xa9, 0xef,
	0x90, 0x36, 0x98, 0x2d, 0xe5, 0xaf, 0x85, 0x8f, 0xa1, 0x85, 0x33, 0xe9, 0x41, 0xf3, 0x8d, 0xf9,
	0x57, 0x0c, 0x7a, 0x4c, 0x87, 0xbe, 0x43, 0x5c, 0xd8, 0x38, 0xa6, 0xc3, 0xc3, 0x87, 0xfe, 0x9a,
	0x89, 0xbd, 0x7e, 0x74, 0xe8, 0xaf, 0x13, 0x80, 0xcd, 0xd7, 0x8f, 0x0e, 0xa3, 0xc3, 0x87, 0x7e,
	0x2b, 0xbc, 0x00, 0x6f, 0xa9, 0x05, 0x66, 0x89, 0x57, 0x8a, 0x47, 0x17, 0x22, 0x67, 0xb8, 0xea,
	0x3b, 0xb4, 0x5d, 0x29, 0xfe, 0x54, 0xe4, 0xcc, 0xcc, 0xbc, 0x81, 0xe4, 0x84, 0xe3, 0x7a, 0xef,
	0xd0, 0xcd, 0x4a, 0x71, 0x3a, 0xe1, 0xe4, 0x2b, 0xe8, 0x4d, 0x85, 0xe9, 0xc1, 0x5c, 0xb9, 0x8e,
	0xf8, 0x16, 0x46, 0xc7, 0x56, 0x1e, 0x0a, 0x20, 0x1f, 0xb7, 0x80, 0xf4, 0xe1, 0x1a, 0xce, 0x42,
	0x54, 0x5e, 0xce, 0x54, 0x1a, 0xb3, 0x2c, 0xca, 0x79, 0x2e, 0xe4, 0x0c, 0x0f, 0x6f, 0xd1, 0x1d,
	0x04, 0x5f, 0xd4, 0xd8, 0x10, 0x21, 0x73, 0x23, 0xb0, 0xb7, 0x2c, 0xcd, 0xd8, 0x24, 0xe3, 0x66,
	0x0d, 0x2a, 0xcc, 0x67, 0x83, 0x76, 0xe7, 0xd1, 0x41, 0x59, 0xa9, 0xf0, 0x77, 0x07, 0x3a, 0xcd,
	0xc4, 0x11, 0x02, 0xad, 0x84, 0xab, 0x18, 0x6d, 0x5d, 0x8a, 0xbf, 0x4d, 0x0c, 0xdf, 0xae, 0xbd,
	0xac, 0xf0, 0x37, 0xb9, 0x0d, 0xa0, 0x34, 0x93, 0x1a, 0x6f, 0x3c, 0xac, 0xa3, 0x45, 0x5d, 0x8c,
	0x98, 0x8b, 0x8e, 0xdc, 0x02, 0x57, 0x72, 0x96, 0x59, 0xb4, 0x85, 0x68, 0xc7, 0x04, 0x10, 0xbc,
	0x0d, 0x60, 0x93, 0x37, 0x8d, 0xc0, 0x8b, 0xa7, 0x45, 0x5d, 0x1b, 0x19, 0x2b, 0x1e, 0xfe, 0xe9,
	0x40, 0x6f, 0x28, 0x92, 0x2a, 0xe3, 0x2f, 0x67, 0xa5, 0xad, 0xfe, 0xd7, 0x66, 0x50, 0xed, 0x20,
	0x60, 0x76, 0xbd, 0xfe, 0xfd, 0xd5, 0x1b, 0xf5, 0x03, 0xa9, 0x9d, 0x5b, 0xdb, 0xd0, 0xa5, 0xdd,
	0x3a, 0x59, 0x44, 0xc9, 0xff, 0xc1, 0xcb, 0x51, 0x13, 0xe9, 0x59, 0xd9, 0x54, 0x09, 0xf9, 0xdc,
	0xc6, 0xbc, 0xb7, 0xa2, 0xca, 0x23, 0x31, 0x8d, 0x6c, 0x50, 0x61, 0xbd, 0x5d, 0xba, 0x55, 0x54,
	0xf9, 0x68, 0x6a, 0xcf, 0x53, 0xe1, 0xfd, 0x7a, 0x40, 0x6a, 0xd7, 0x0f, 0xa6, 0xcc, 0x85, 0x8d,
	0xf3, 0xd1, 0xe8, 0xcc, 0x8c, 0x63, 0x07, 0x5a, 0xc3, 0xe3, 0xe7, 0xa7, 0xfe, 0x5a, 0x98, 0xc1,
	0xcd, 0x81, 0x4c, 0xb5, 0x79, 0x61, 0x63, 0xc5, 0xe5, 0x8f, 0xa2, 0x92, 0x05, 0x9f, 0x35, 0xdf,
	0x66, 0xd3, 0x74, 0x67, 0xa9, 0xe9, 0x47, 0xd0, 0x6e, 0xbe, 0xfd, 0xb5, 0xff, 0xf8, 0x54, 0x97,
	0xee, 0x14, 0xda, 0x08, 0xc2, 0x09, 0xdc, 0x5a, 0x71, 0x9a, 0x5a, 0xac, 0x82, 0x56, 0x5c, 0xbd,
	0x51, 0x81, 0x83, 0xfb, 0x6c, 0x75, 0x67, 0xff, 0x3d, 0x5b, 0x8a, 0xe2, 0xf0, 0x0f, 0x07, 0xb6,
	0x3f, 0x5a, 0x3c, 0x24, 0x80, 0x76, 0xd3, 0x37, 0x07, 0xfb, 0xd6, 0x3c, 0x9a, 0xd5, 0x51, 0xdf,
	0xcc, 0xb6, 0xa0, 0x2e, 0x9d, 0x3f, 0x93, 0x6f, 0x60, 0xdb, 0x0e, 0x3c, 0xcb, 0x32, 0x11, 0x47,
	0xb1, 0xa8, 0x0a, 0x5d, 0xcf, 0xd9, 0x55, 0x04, 0x8e, 0x4d, 0x7c, 0x60, 0xc2, 0x64, 0x1f, 0xfc,
	0x65, 0xae, 0x4a, 0xdf, 0x37, 0x43, 0xd7, 0x5b, 0x50, 0xcf, 0xd3, 0xf7, 0xdc, 0x5c, 0x85, 0x39,
	0x7b, 0x17, 0x5d, 0x72, 0x56, 0x5a, 0x9a, 0x9d, 0x3e, 0x2f, 0x67, 0xef, 0x7e, 0xe0, 0xac, 0x34,
	0x9c, 0x93, 0x6b, 0xbf, 0xd4, 0xdb, 0xb6, 0xae, 0x3b, 0xc2, 0x7f, 0x83, 0x7f, 0x07, 0x00, 0x00,
	0xff, 0xff, 0xa3, 0xdd, 0xbb, 0xb3, 0x1d, 0x0a, 0x00, 0x00,
}
